{*
 * Basic structure of a FFmpeg-based audio/video player for FreePascal (FPC)
 *
 * Dependencies: FFmpeg v4.0.2 and SDL v2
 *
 * Copyright (C) 2020 Robert Bakos <av4all@gmx.com>
 *
 * This file is part of project FFPlay4Laz
 * <http://forum.lazarus.freepascal.org/index.php/topic,26666.0.html>
 *
 * based on Dranger-Tutorial
 *          <http://dranger.com/ffmpeg/>
 *          and
 *          ffexample.c by Lovesan
 *          <https://gist.github.com/Lovesan/e29d815586ea20642d85875d47cefe03>
 *
 *
 * The original FFmpeg library license applies if you use the FFmpeg libraries.
 * See <https://www.ffmpeg.org/legal.html> for license details.
 *
 * This source is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * A copy of the GNU General Public License is available on the World Wide Web
 * at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
 * to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
 * Boston, MA 02110-1335, USA.
 *
 *}


program FFPlay4Laz2_MINIMALST;

{$mode objfpc}{$H+}

uses
  sdl2, SysUtils,
  { FFmpeg... }
  libavformat, libavcodec,
  libavutil, libavutil_channel_layout, libavutil_frame, libavutil_imgutils, libavutil_mem,
             libavutil_pixfmt, libavutil_rational, libavutil_samplefmt, libavutil_time,
  libswresample, libswscale;

const
  FF_ALLOC_EVENT   = (SDL_USEREVENT);
  FF_REFRESH_EVENT = (SDL_USEREVENT + 1);
  FF_QUIT_EVENT    = (SDL_USEREVENT + 2);

  MAX_PACKETQUEUE_SIZE  = (15 * 1024 * 1024);
  MAX_PICTUREQUEUE_SIZE = 2; {$<- Note: Two alternating VideoPictures are used. }

  SDL_AUDIOBUFFER_SIZE = 1024;
  MAX_AUDIOBUFFER_SIZE = (192000 * 3) div 2;

  AV_SYNC_THRESHOLD_MIN = 0.01;
  AV_NOSYNC_THRESHOLD   = 10.0;

type
  { PacketQueues }
  TPacketQueue = record
    FirstPkt, LastPkt: PAVPacketList;
    NmbPackets, Size : Integer;
    SdlMutex: PSDL_Mutex;
    SdlCond : PSDL_Cond;
  end;

  { SDL-VideoImage }
  PVideoPicture = ^TVideoPicture;
  TVideoPicture = record
    SdlImg       : PSDL_Texture;
    Width, Height: Integer;
    IsAllocated  : Boolean;
    pts          : Double;
  end;

  { FFPlayer... }
  PffPlayer = ^TffPlayer;
  TffPlayer = record
    FormatCtx             : PAVFormatContext;
    CodecCtxA , CodecCtxV : PAVCodecContext;
    StreamA   , StreamV   : PAVStream;
    StreamAIdx, StreamVIdx: Integer;
    PktQA     , PktQV     : TPacketQueue;
    AudioBuf  , AudioBufRs: PByte;
    FrameVOut , FrameA    : PAVFrame;
    AudioBufSize, AudioBufPos: Integer;

    { FFAudioResampler }
    WantedSmpFmt   : TAVSampleFormat;
    WantedSmpRate  : Integer;
    WantedChannels : Integer;
    WantedChnLayout: Int64;
    TgtFrameSize   : Integer;
    TgtDataSize    : Integer;
    SwrCtx: PSwrContext;

    { FFVideoRescaler/Converter }
    WantedFFPxlFmt: TAVPixelFormat;
    SwsCtx: PSwsContext;

    { PictureQueue }
    picQ: array[0..MAX_PICTUREQUEUE_SIZE-1] of TVideoPicture;
    picQ_Size, picQ_rIdx, picQ_wIdx: Byte;  // PictureQueue-Size/ReadIndex/WriteIndex
    picQ_SdlMutex: PSDL_Mutex;
    picQ_SdlCond : PSDL_Cond;

    { FFPlayer-Threads }
    ReadThreadID : PSDL_Thread;
    VideoThreadID: PSDL_Thread;

    { SDL-Window }
    SdlWindow      : PSDL_Window;
    SdlRenderer    : PSDL_Renderer;
    WantedSdlPxlFmt: Cardinal;

    { AV-Sync }
    ClkASec, ClkVSec, LastFrameVPts, LastFrameVDelay, FrameVTimerSec: Double;

    QuitRead: Boolean;
  end;

var
  Pff: PffPlayer;
  Tff: TffPlayer;


{ LogMessages }
procedure Log(const aLogMsg: String);
begin
  WriteLn(aLogMsg);
end;

{--- PacketQueues -------------------------------------------------------------}

{ Init a PacketQueue }
procedure pktQ_Init(var aPktQ: TPacketQueue);
begin
  FillChar(aPktQ, SizeOf(TPacketQueue), 0);  // Fill the PacketQueue with ZEROs

  aPktQ.SdlMutex := SDL_CreateMutex();
  aPktQ.SdlCond  := SDL_CreateCond();
end;

{ Put an AV-Packet into its PacketQueue }
function pktQ_Put(var aPktQ: TPacketQueue; const aPkt: PAVPacket): Integer;
var
  newPkt: PAVPacketList;
begin
  newPkt := av_malloc(SizeOf(TAVPacketList));
  if (newPkt = nil) then Exit(-1);

  newPkt^.pkt  := aPkt^;
  newPkt^.next := nil;

  SDL_LockMutex(aPktQ.SdlMutex);

  if (aPktQ.LastPkt = nil) then
    aPktQ.FirstPkt := newPkt
  else
    aPktQ.LastPkt^.next := newPkt;

  aPktQ.LastPkt    := newPkt;
  aPktQ.NmbPackets += 1;
  aPktQ.Size       += newPkt^.pkt.size;

  SDL_CondSignal(aPktQ.SdlCond);
  SDL_UnlockMutex(aPktQ.SdlMutex);

  Result := 0; {0=Success}
end;

{ Get an AV-Packet from its PacketQueue }
function pktQ_Get(var aPktQ: TPacketQueue; var aPkt: TAVPacket; const aDoBlock: Boolean): Integer;
var
  aFirstPkt: PAVPacketList;
  aRet: Integer;
begin
  Result := -1; {<0=aborted} // PRESET

  SDL_LockMutex(aPktQ.SdlMutex);

  while True do
  begin

    if Pff^.QuitRead then
    begin
      aRet := -1; {<0=aborted}
      Break;
    end;

    aFirstPkt := aPktQ.FirstPkt;
    if (aFirstPkt <> nil) then
    begin
      aPktQ.FirstPkt := aFirstPkt^.next;
      if (aPktQ.FirstPkt = nil) then
        aPktQ.LastPkt := nil;
      aPktQ.NmbPackets -= 1;
      aPktQ.Size       -= aFirstPkt^.pkt.size;
      aPkt := aFirstPkt^.pkt;

      av_free(aFirstPkt);

      aRet := 1; {>0=Pkt was FETCHED}
      Break;
    end

    else if not aDoBlock then
    begin
      aRet := 0; {0=NO Pkt fetched}
      Break;
    end

    else
      SDL_CondWait(aPktQ.SdlCond, aPktQ.SdlMutex);

  end;

  SDL_UnlockMutex(aPktQ.SdlMutex);

  Result := aRet;
end;


{--- AUDIO --------------------------------------------------------------------}

{ AudioClock in Sec. }
function clk_GetClkASec(): Double;
var
  aClkA: Double;
  aHWBufSize, aBytesPerSec: Integer;
begin

  aClkA := Pff^.ClkASec;

  aHWBufSize := Pff^.AudioBufSize - Pff^.AudioBufPos;
  aBytesPerSec := 0;

  if (Pff^.CodecCtxA <> nil) then
    aBytesPerSec := Pff^.TgtFrameSize * Pff^.WantedSmpRate;
  if (aBytesPerSec > 0) then
    aClkA -= (aHWBufSize / aBytesPerSec);

  Result := aClkA;
end;


{ AudioDecode }
function audio_DecodeFrame(): Integer;
var
  aRet       : Integer;
  aPktA      : TAVPacket;
  aHasPktA   : Boolean = False;
  aHasFramesA: Boolean = False;
begin
  Result := -1;  // PRESET

  FillChar(aPktA, SizeOf(TAVPacket), 0);  // Fill the AUDIO-Packet with ZEROs

  { Allocate the AudioFrame-Buffer }
  Pff^.FrameA := av_frame_alloc();
  if (Pff^.FrameA = nil) then
  begin
    Log('! FF-Err: Cannot allocate the AudioFrame-Buffer.');
    Exit(-1);
  end;


  { Decode AudioFrames... }

  while True do
  begin

    while (aHasFramesA) do  {<- Note: The AudioPacket can contain SEVERAL Frames. }
    begin

      aRet := avcodec_receive_frame(Pff^.CodecCtxA, Pff^.FrameA);
      if (aRet <> 0) then
      begin
        aHasFramesA := False;
        Break;
      end;

      { Get DataSize for the WANTED AudioFormat (= the TARGET-Datasize) }
      Pff^.TgtDataSize := av_samples_get_buffer_size(nil,
                                                     Pff^.WantedChannels,
                                                     Pff^.FrameA^.nb_samples,
                                                     Pff^.WantedSmpFmt,
                                                     1);

      { Allocate&Clear Memory for the FFAudioResampler-Buffer }
      if (Pff^.AudioBufRs = nil) then
        Pff^.AudioBufRs := AllocMem({Pff^.TgtDataSize}MAX_AUDIOBUFFER_SIZE);

      { Convert to WANTED AudioFormat (= the TARGET-AudioFormat) }
      swr_convert(Pff^.SwrCtx,
          {out->} @Pff^.AudioBufRs     , Pff^.FrameA^.nb_samples,
          { in->} @Pff^.FrameA^.data[0], Pff^.FrameA^.nb_samples);

      { Move CONVERTED AudioData to the AudioBuffer }
      Move(Pff^.AudioBufRs^, Pff^.AudioBuf^, Pff^.TgtDataSize);

      { UPDATE the AudioClock, in Case of there's NO PTS }
      Pff^.ClkASec += Pff^.TgtDataSize / (Pff^.TgtFrameSize * Pff^.WantedSmpRate);
      Exit(Pff^.TgtDataSize);
    end;


    { RESET the current AUDIO-Packet }
    if aHasPktA then
      av_packet_unref(@aPktA);

    if Pff^.QuitRead then Exit(-1);

    { Get the NEXT AUDIO-Packet from its PacketQueue }
    if ( pktQ_Get(Pff^.PktQA, aPktA, True) < 0 ) then
      Exit(-1);
    aHasPktA := True;

    { IF there's a PTS...UPDATE the AudioClock with it }
    if (aPktA.pts <> AV_NOPTS_VALUE) then
      Pff^.ClkASec := aPktA.pts * av_q2d(Pff^.StreamA^.time_base);

    aRet := avcodec_send_packet(Pff^.CodecCtxA, @aPktA);
    if (aRet <> 0) then
    begin
      Log('! FF-Err: Cannot send AUDIO-Packet to the Decoder.');
      Exit(aRet);
    end;

    aHasFramesA := True;
  end;

  Result := -1; {-1=Error -> Output Silence}
end;


{ AudioCallback }
procedure audio_CB(aUserData: Pointer; aStream: PUInt8; aLen: Integer); cdecl;
var
  nextLenA, aAudioSize: Integer;
  nextBufA: PByte;
begin

  while (aLen > 0) do
  begin

    if (Pff^.AudioBufPos >= Pff^.AudioBufSize) then
    begin
      aAudioSize := audio_DecodeFrame();

      { OnError...Output Silence }
      if (aAudioSize < 0) then
      begin
        Pff^.AudioBufSize := SDL_AUDIOBUFFER_SIZE;
        FillChar(Pff^.AudioBuf^, Pff^.AudioBufSize, 0);
      end
      else
        Pff^.AudioBufSize := aAudioSize;

      Pff^.AudioBufPos := 0;
    end;

    nextLenA := Pff^.AudioBufSize - Pff^.AudioBufPos;
    if (nextLenA > aLen) then nextLenA := aLen;

    { Move AudioData to the AudioOutput }
    nextBufA := Pff^.AudioBuf + Pff^.AudioBufPos;
    Move(nextBufA^, aStream^, nextLenA);

    aLen             -= nextLenA;
    aStream          += nextLenA;
    Pff^.AudioBufPos += nextLenA;

  end;
end;


{--- VIDEO --------------------------------------------------------------------}

{ Allocate a NEW SDL-VideoImage in the PictureQueue }
procedure picQ_AllocPic();
var
  newPic: PVideoPicture;
begin
  newPic := @Tff.picQ[Pff^.picQ_wIdx];

  { Create a NEW SDL-VideoImage... }
  if (newPic^.SdlImg <> nil) then
    SDL_DestroyTexture(newPic^.SdlImg);
  newPic^.SdlImg := SDL_CreateTexture(Pff^.SdlRenderer,
                                      Pff^.WantedSdlPxlFmt,
                                      SDL_TEXTUREACCESS_STREAMING,
                                      Pff^.CodecCtxV^.width, Pff^.CodecCtxV^.height);
  { Set the NEW SDL-VideoImage's Size }
  newPic^.width  := Pff^.CodecCtxV^.width;
  newPic^.height := Pff^.CodecCtxV^.height;

  SDL_LockMutex(Pff^.picQ_SdlMutex);
  newPic^.IsAllocated := True;
  SDL_CondSignal(Pff^.picQ_SdlCond);
  SDL_UnlockMutex(Pff^.picQ_SdlMutex);
end;

{ Put the NEXT VideoFrame into the PictureQueue (= Queue Picture) }
function picQ_QueuePic(const aFrameVIn: PAVFrame; const aFrameVPts: Double): Integer;
var
  qPic: PVideoPicture;
  aSdlEvent: TSDL_event;
begin
  Result := -1;  // PRESET

  { IF the PictureQueue is too BIG...WAIT }
  SDL_LockMutex(Pff^.picQ_SdlMutex);
  while ( (Pff^.picQ_Size >= MAX_PICTUREQUEUE_SIZE) and (not Pff^.QuitRead) ) do
    SDL_CondWait(Pff^.picQ_SdlCond, Pff^.picQ_SdlMutex);
  SDL_UnlockMutex(Pff^.picQ_SdlMutex);

  if Pff^.QuitRead then Exit(-1);

  { Set the NEXT Pic to be QUEUED in the PictureQueue (= Queue Pic) }
  qPic := @Tff.picQ[Pff^.picQ_wIdx];

  { Allocate or Resize the NEXT Pic... }
  if ((qPic^.SdlImg = nil)                      or
      (qPic^.Width  <> Pff^.CodecCtxV^.width)   or
      (qPic^.Height <> Pff^.CodecCtxV^.height)) then
  begin
    qPic^.IsAllocated := False;  // RESET

    aSdlEvent.type_ := FF_ALLOC_EVENT;
    SDL_PushEvent(@aSdlEvent);

    { WAIT, until the NEXT Pic IS allocated }
    SDL_LockMutex(Pff^.picQ_SdlMutex);
    while ( (not qPic^.IsAllocated) and (not Pff^.QuitRead) ) do
      SDL_CondWait(Pff^.picQ_SdlCond, Pff^.picQ_SdlMutex);
    SDL_UnlockMutex(Pff^.picQ_SdlMutex);

    if Pff^.QuitRead then Exit(-1);
  end;

  if (qPic^.SdlImg <> nil) then
  begin
    { Scale/Convert the VideoFrame }
    sws_scale(Pff^.SwsCtx,
      { in->} aFrameVIn^.data, aFrameVIn^.linesize, 0, Pff^.CodecCtxV^.height,
      {out->} Pff^.FrameVOut^.data, Pff^.FrameVOut^.linesize);

    { UPDATE the SDL-VideoImage }
    SDL_UpdateTexture(qPic^.SdlImg, nil, Pff^.FrameVOut^.data[0], Pff^.FrameVOut^.linesize[0]);
    qPic^.pts := aFrameVPts;

    { UPDATE the PictureQueue... }
    Pff^.picQ_wIdx += 1;
    if (Pff^.picQ_wIdx = MAX_PICTUREQUEUE_SIZE) then Pff^.picQ_wIdx := 0;  // RESET the WriteIndex

    SDL_LockMutex(Pff^.picQ_SdlMutex);
    Pff^.picQ_Size += 1;
    SDL_UnlockMutex(Pff^.picQ_SdlMutex);
  end;

  Result := 0; {0=Success}
end;


{ Sync Video }
function syncVideo(const aFrameV: PAVFrame; aFrameVPts: Double): Double;
var
  aFrameVDelaySec: Double;
begin

  if (aFrameVPts <> 0) then
    Pff^.ClkVSec := aFrameVPts
  else
    aFrameVPts := Pff^.ClkVSec;

  aFrameVDelaySec := av_q2d(Pff^.StreamV^.time_base);
  aFrameVDelaySec += aFrameV^.repeat_pict * (aFrameVDelaySec * 0.5);
  Pff^.ClkVSec += aFrameVDelaySec;

  Result := aFrameVPts;
end;


{--- VIDEO-Thread -------------------------------------------------------------}

function thread_Video(aArg: Pointer): Integer; cdecl;
var
  aPktV     : TAVPacket;
  aFrameV   : PAVFrame = nil;
  aFrameVPts: Double;
  aRet: Integer;
begin
  Result := -1;  // PRESET

  { Allocate the VideoFrame-Buffer }
  aFrameV := av_frame_alloc();
  if (aFrameV = nil) then
  begin
    Log('! FF-Err: Cannot allocate the VideoFrame-Buffer.');
    Exit(-1);
  end;


  { Decode VideoFrames... }

  { Get the NEXT VIDEO-Packet from its PacketQueue... }
  while ( pktQ_Get(Pff^.PktQV, aPktV, True) > 0 ) do
  begin

    aRet := avcodec_send_packet(Pff^.CodecCtxV, @aPktV);
    if (aRet < 0) then
    begin
      Log('! FF-Err: Cannot send VIDEO-Packet to the Decoder.');
      Continue;
    end;

    while (avcodec_receive_frame(Pff^.CodecCtxV, aFrameV) = 0) do
    begin
      { Get the DECODED VideoFrame's PTS }
      if (aPktV.dts <> AV_NOPTS_VALUE) then
        aFrameVPts := av_frame_get_best_effort_timestamp(aFrameV)
      else
        aFrameVPts := 0;

      aFrameVPts *= av_q2d(Pff^.StreamV^.time_base);
      aFrameVPts := syncVideo(aFrameV, aFrameVPts);

      { Put the VideoFrame into the PictureQueue }
      if ( picQ_QueuePic(aFrameV, aFrameVPts) < 0 ) then
        Break;
    end;

    av_packet_unref(@aPktV);

  end;

  av_frame_free(@aFrameV);

  Result := 0; {0=Success}
end;


{--- READ-Thread --------------------------------------------------------------}

function thread_Read(aArg: Pointer): Integer; cdecl;
var
  aPkt: TAVPacket;
  aSdlEvent: TSDL_Event;
begin
  Result := -1;  // PRESET

  Log(sLineBreak + '> Opened PLAYING - To stop, press <Esc>; to exit click [X].');

  while True do
  begin
    if Pff^.QuitRead then Break;

    { LIMIT the PacketQueues }
    if ((Pff^.PktQA.Size > MAX_PACKETQUEUE_SIZE)  or
        (Pff^.PktQV.Size > MAX_PACKETQUEUE_SIZE)) then
    begin
      SDL_Delay(10);
      Continue;
    end;

    { Read AV-Frames }
    if ( av_read_frame(Pff^.FormatCtx, @aPkt) < 0 ) then
      if (Pff^.FormatCtx^.pb^.error = 0) then {0 means EOF reached}
        Break
      else
      begin
        Log('! FF-Err: Cannot read AV-Frame.');
        Continue;
      end;

    { Put the AV-Packet into its PacketQueue }
    if      (aPkt.stream_index = Pff^.StreamAIdx) then
      pktQ_Put(Pff^.PktQA, @aPkt)
    else if (aPkt.stream_index = Pff^.StreamVIdx) then
      pktQ_Put(Pff^.PktQV, @aPkt)
    else
      av_packet_unref(@aPkt);  // RESET it

  end;
  if not Pff^.QuitRead then  // IF NOT has been stopped via <Esc>
    Log('> EoF reached -> Playing from the PacketQueues.');

  { Wait for [X] }
  while (not Pff^.QuitRead) do
    SDL_Delay(100);

  aSdlEvent.type_ := FF_QUIT_EVENT;
  SDL_PushEvent(@aSdlEvent);

  Result := 0; {0=Success}
end;


{--- Sync/Show VIDEO ----------------------------------------------------------}

{ SDL-VideoRefresh-TimerCallback }
function video_Refresh_CB(aInterval: uInt32; aParam: Pointer): uInt32; cdecl;
var
  aSdlEvent: TSDL_event;
begin
  aSdlEvent.type_ := FF_REFRESH_EVENT;
  SDL_PushEvent(@aSdlEvent);

  Result := 0; {$<- 0 cancels the SDL-Timer }
end;

{ SDL-VideoRefresh-Timer }
procedure video_ScheduleRefresh(const aDelayVMs: Integer);
begin
  SDL_AddTimer(aDelayVMs, @video_Refresh_CB, nil);
end;          {'current VideoRefresh-Delay in Ms }


{ Display the current Picture }
procedure video_ShowPic(const showPic: PVideoPicture);
var
  aSdlImgRect : TSDL_Rect;
  aAspectRatio: Single;
  w, h, x, y, aSdlWndW, aSdlWndH: Integer;
begin

  { Get the AspectRatio... }
  if (Pff^.CodecCtxV^.sample_aspect_ratio.num = 0) then {= SAR}
    aAspectRatio := 0
  else
    aAspectRatio := av_q2d(Pff^.CodecCtxV^.sample_aspect_ratio) *
                          (Pff^.CodecCtxV^.width / Pff^.CodecCtxV^.height); {= DAR}
  if (aAspectRatio <= 0.0) then
    aAspectRatio := Pff^.CodecCtxV^.width / Pff^.CodecCtxV^.height;

  { Get the SDL-VideoImage's Pos&Size... }
  SDL_GetWindowSize(Pff^.SdlWindow, @aSdlWndW, @aSdlWndH);
  h := aSdlWndH;
  w := Round(h * aAspectRatio) and -3;
  if (w > aSdlWndW) then
  begin
    w := aSdlWndW;
    h := Round(w / aAspectRatio) and -3;
  end;

  x := Round((aSdlWndW - w) / 2);
  y := Round((aSdlWndH - h) / 2);
  aSdlImgRect.x := x;
  aSdlImgRect.y := y;
  aSdlImgRect.w := w;
  aSdlImgRect.h := h;

  { Display the SDL-VideoImage }
  SDL_RenderClear  (Pff^.SdlRenderer);
  SDL_RenderCopy   (Pff^.SdlRenderer, showPic^.SdlImg, nil, @aSdlImgRect);
  SDL_RenderPresent(Pff^.SdlRenderer);
end;


{ Refresh the SDL-VideoImage }
procedure video_Refresh();
var
  dqPic: PVideoPicture;
  nextDelayV, realDelayV, aSyncVThreshold, aDiffVA: Double;
begin

  if (Pff^.CodecCtxV <> nil) then
  begin

    if (Pff^.picQ_Size = 0) then
      video_ScheduleRefresh(10)

    else
    begin
      { Get the NEXT Pic to be DISPLAYED from the PictureQueue (= DeQueue Pic) }
      dqPic := @Tff.picQ[Pff^.picQ_rIdx];

      { Calculate the NEXT VideoDelay... }
      {$'Note: This is a BASIC AV-Sync, that suffices for general needs. }
      {$       For more accurate AV-SyncCodes, see 'ffplay.c' by FFmpeg. }

      nextDelayV := dqPic^.pts - Pff^.LastFrameVPts;
      { IF NOT VALID...Use the LAST one }
      if ( (nextDelayV <= 0) or (nextDelayV >= 1.0) ) then
        nextDelayV := Pff^.LastFrameVDelay;

      { STORE for next Time }
      Pff^.LastFrameVDelay := nextDelayV;
      Pff^.LastFrameVPts   := dqPic^.pts;

      aDiffVA := dqPic^.pts - clk_GetClkASec();

      if (nextDelayV > AV_SYNC_THRESHOLD_MIN) then
        aSyncVThreshold := nextDelayV
      else
        aSyncVThreshold := AV_SYNC_THRESHOLD_MIN;

      if (Abs(aDiffVA) < AV_NOSYNC_THRESHOLD) then
        if (aDiffVA <= -aSyncVThreshold) then
          nextDelayV := 0
        else if (aDiffVA >= aSyncVThreshold) then
          nextDelayV := 2 * nextDelayV;

      { UPDATE the VideoFrame-Timer in Sec. }
      Pff^.FrameVTimerSec += nextDelayV;

      realDelayV := Pff^.FrameVTimerSec - (av_gettime() / 1000000.0);
      if (realDelayV < AV_SYNC_THRESHOLD_MIN) then
        realDelayV := AV_SYNC_THRESHOLD_MIN; {<- Note: The Pic should be skipped instead. }

      video_ScheduleRefresh( Round(realDelayV*1000 + 0.5) );

      { Display the current Picture }
      video_ShowPic(dqPic);


      { UPDATE the PictureQueue... }
      Pff^.picQ_rIdx += 1;
      if (Pff^.picQ_rIdx = MAX_PICTUREQUEUE_SIZE) then Pff^.picQ_rIdx := 0;  // RESET the ReadIndex

      SDL_LockMutex(Pff^.picQ_SdlMutex);
      Pff^.picQ_Size -= 1;
      SDL_CondSignal (Pff^.picQ_SdlCond);
      SDL_UnlockMutex(Pff^.picQ_SdlMutex);
    end;

  end
  else
    video_ScheduleRefresh(100);
end;


{--- Open ---------------------------------------------------------------------}

{ Open the FFPlayer-Components }
function ff_OpenComponents(const aStreamIdx: Integer; const aFormatCtx: PAVFormatContext): Integer;
var
  aCodecParams : PAVCodecParameters;
  aCodec       : PAVCodec = nil;
  aCodecCtx    : PAVCodecContext = nil;
  aWantedSpec, aAudioSpec: TSDL_AudioSpec;
  aFrameVIn    : PByte;
  aFrameVInSize: Integer;
begin
  Result := -1;  // PRESET

  { Check, IF the StreamIndex is VALID }
  if ( (aStreamIdx < 0) or (aStreamIdx >= aFormatCtx^.nb_streams) ) then
    Exit(-1);

  { Get the Codec (-> the Decoder) }
  aCodecParams := aFormatCtx^.streams[aStreamIdx]^.codecpar;  // Get the CodecParams
  aCodec := avcodec_find_decoder(aCodecParams^.codec_id);
  if (aCodec = nil) then
  begin
    Log('! Codec not supported.');
    Exit(-1);
  end
  else
    Log(sLineBreak + '> Codec found: ' + aCodec^.name + ' = ' + aCodec^.long_name);

  { Init the AV-CodecContext... }

  aCodecCtx := avcodec_alloc_context3(aCodec);
  if (aCodecCtx = nil) then Exit(-1);

  if ( avcodec_parameters_to_context(aCodecCtx, aCodecParams) < 0 ) then
  begin
    avcodec_free_context(@aCodecCtx);
    Exit(-1);
  end;

  if ( avcodec_open2(aCodecCtx, aCodec, nil) < 0 ) then
  begin
    Log('! Codec not supported.');
    avcodec_free_context(@aCodecCtx);
    Exit(-1);
  end;


  { Open the AUDIO-Component... }

  if (aCodecParams^.codec_type = AVMEDIA_TYPE_AUDIO) then
  begin
    Pff^.CodecCtxA    := aCodecCtx;
    Pff^.StreamAIdx   := aStreamIdx;
    Pff^.AudioBufSize := 0;  // PRESET
    Pff^.AudioBufPos  := 0;  // PRESET
    Pff^.StreamA := aFormatCtx^.streams[aStreamIdx];

    { Set&Init the FFAudioResampler... }

    { IF there's NO ChannelLayout...Get the DEFAULT-ChannelLayout } {sometimes needed}
    if (aCodecCtx^.channel_layout = 0) then
      aCodecCtx^.channel_layout := av_get_default_channel_layout(aCodecCtx^.channels);

    {$ Set WANTED AudioFormat, e.g. 'Float32, stereo'... }
    Pff^.WantedSmpFmt    :={aCodecCtx^.sample_fmt     ->}AV_SAMPLE_FMT_FLT;
    Pff^.WantedSmpRate   := aCodecCtx^.sample_rate   {-> 48000}; {<- With slow PCs better done by the AudioOutput }
    Pff^.WantedChannels  :={aCodecCtx^.channels       ->}2;
    Pff^.WantedChnLayout :={aCodecCtx^.channel_layout ->}AV_CH_LAYOUT_STEREO_DOWNMIX{AV_CH_LAYOUT_STEREO};

    { Get the TARGET-AudioFrameSize }
    Pff^.TgtFrameSize := av_get_bytes_per_sample(Pff^.WantedSmpFmt) * Pff^.WantedChannels;

    { Allocate&Set the FFAudioResampler-Context }
    Pff^.SwrCtx := swr_alloc_set_opts(nil,
                              {out->} Pff^.WantedChnLayout     , Pff^.WantedSmpFmt    , Pff^.WantedSmpRate,
                              { in->} aCodecCtx^.channel_layout, aCodecCtx^.sample_fmt, aCodecCtx^.sample_rate,
                                      0, nil);

    { Init the FFAudioResampler-Context }
    if ( swr_init(Pff^.SwrCtx) < 0 ) then
    begin
      Log('! FF-Err: Cannot initialize the FFAudioResampler-Context.');
      Exit(-1);
    end;


    { Open SDL-Audio-Output... }

    aWantedSpec.channels := Pff^.WantedChannels;
    aWantedSpec.freq     := Pff^.WantedSmpRate;
    aWantedSpec.format   := AUDIO_F32; {Float32}
    aWantedSpec.silence  := 0;
    aWantedSpec.samples  := SDL_AUDIOBUFFER_SIZE;
    aWantedSpec.callback := @audio_CB;

    if ( SDL_OpenAudio(@aWantedSpec, @aAudioSpec) = -1 ) then
    begin              {'= DESIRED    '= OBTAINED SDL-AudioParams }
      Log( Format('! SDL-Err: Cannot open SDL-Audio (SDL-Err: %s).', [SDL_GetError]) );
      Exit(-1);
    end
    else
      Log( '> SDL-Audio opened (= always converted to Float32, stereo)'
         + sLineBreak
         + '  SDL-AudioDriver: ' + SDL_GetCurrentAudioDriver()
         );

    { Init the AUDIO-PacketQueue }
    pktQ_Init(Pff^.PktQA);

    { Allocate&Clear Memory for the AudioBuffer }
    Pff^.AudioBuf := AllocMem(MAX_AUDIOBUFFER_SIZE);

    { Start the SDL-AudioDevice PLAYING }
    SDL_PauseAudio(0);
  end


  { Open the VIDEO-Component... }

  else if (aCodecParams^.codec_type = AVMEDIA_TYPE_VIDEO) then
  begin
    Pff^.CodecCtxV  := aCodecCtx;
    Pff^.StreamVIdx := aStreamIdx;
    Pff^.StreamV    := aFormatCtx^.streams[aStreamIdx];

    Pff^.FrameVTimerSec  := av_gettime() / 1000000.0;
    Pff^.LastFrameVDelay := 40e-3; // PRESET = 40Ms = 25 FPS

    {$ Set WANTED VideoFormat... }
    Pff^.WantedFFPxlFmt  := AV_PIX_FMT_YUV420P;   {YUV} {faster -> better for slow PCs}
    Pff^.WantedSdlPxlFmt := SDL_PIXELFORMAT_IYUV;
    Log('  PixelFormat: YUV 12bpp (faster)');
(*
    Pff^.WantedFFPxlFmt  := AV_PIX_FMT_RGB24;     {RGB}
    Pff^.WantedSdlPxlFmt := SDL_PIXELFORMAT_RGB24;
    Log('  PixelFormat: RGB 24bpp (nicer)');
*)
    { Allocate the FFVideoRescaler/Converter-Context }
    Pff^.SwsCtx := sws_getContext(aCodecCtx^.width, aCodecCtx^.height, aCodecCtx^.pix_fmt,
                                  aCodecCtx^.width, aCodecCtx^.height, Pff^.WantedFFPxlFmt,
                                  SWS_BILINEAR, nil, nil, nil);
    if (Pff^.SwsCtx = nil) then
    begin
      Log('! FF-Err: Cannot allocate the FFVideoRescaler/Converter-Context.');
      Exit(-1);
    end;

    { Allocate the VideoFrameOut-Buffer }
    Pff^.FrameVOut := av_frame_alloc();
    if (Pff^.FrameVOut = nil) then
    begin
      Log('! FF-Err: Cannot allocate the VideoFrameOut-Buffer.');
      Exit(-1);
    end;

    aFrameVInSize := av_image_get_buffer_size(Pff^.WantedFFPxlFmt,
                                              aCodecParams^.width, aCodecParams^.height, 8);
    if (aFrameVInSize < 1) then
    begin
      Log('! FF-Err: Cannot get the VideoFrameIn-BufferSize.');
      Exit(-1);
    end;

    { Allocate the VideoFrameIn-Buffer }
    aFrameVIn := av_malloc(aFrameVInSize);
    if (aFrameVIn = nil) then
    begin
      Log('! FF-Err: Cannot allocate the VideoFrameIn-Buffer.');
      Exit(-1);
    end;

    { Fill the VideoImage-Arrays }
    if ( av_image_fill_arrays(@Pff^.FrameVOut^.data[0],
                              @Pff^.FrameVOut^.linesize[0],
                              aFrameVIn,
                              Pff^.WantedFFPxlFmt,
                              aCodecParams^.width,
                              aCodecParams^.height, 1) < 0 ) then
    begin
      Log('! FF-Err: Cannot fill the VideoImage-Arrays.');
      Exit(-1);
    end;


    { Init the VIDEO-PacketQueue }
    pktQ_Init(Pff^.PktQV);

    { Create the VIDEO-Mutexes }
    Pff^.picQ_SdlMutex := SDL_CreateMutex();
    Pff^.picQ_SdlCond  := SDL_CreateCond();

    { ADD the SDL-VideoRefresh-Timer }
    video_ScheduleRefresh(40);
                         {'INITIAL VideoRefresh-Delay = 40Ms = 25 FPS }

    { Start the VIDEO-Thread }
    Pff^.VideoThreadID := SDL_CreateThread(@thread_Video, 'VIDEO-Thread', nil);
    if (Pff^.VideoThreadID = nil) then
    begin
      Log( Format('! SDL-Err: Cannot start the VIDEO-Thread (SDL-Err: %s).', [SDL_GetError]) );
      Exit(-1);
    end;

  end

  else
  begin
    avcodec_free_context(@aCodecCtx);
    Exit(-1);
  end;

  Result := 0; {0=Success}
end;


{*** MAINProcedure ************************************************************}

function ff_main(): Integer;
var
  aURL: String;
  aIdx, aStreamAIdx, aStreamVIdx: Integer;
  aFormatCtx: PAVFormatContext;

  aSdlWndW, aSdlWndH: Integer;
  aSdlWindow  : PSDL_Window   = nil;
  aSdlRenderer: PSDL_Renderer = nil;
  aSdlEvent   : TSDL_Event;

label
  CleanUp;
begin

  { Allocate&Clear the FFPlayer-Memory }
  Pff := av_mallocz(SizeOf(TffPlayer));
  if (Pff = nil) then Exit(-1);

  { Set the URL }
  aURL := ParamStr(1);
  if (aURL = '') then
  begin
    Log(sLineBreak + '! Nothing loaded - Drop a Mediafile on me.' + sLineBreak);
    ExitCode := -1;
    GoTo CleanUp;
  end;

  { Init FFmpeg }
  av_register_all();
  avformat_network_init();

  { Init SDL }
  if ( SDL_Init(SDL_INIT_VIDEO or SDL_INIT_AUDIO or SDL_INIT_TIMER) <> 0 ) then
  begin
    Log( Format('! SDL-Err: Cannot initialize SDL (SDL-Err: %s).', [SDL_GetError]) );
    ExitCode := -1;
    GoTo CleanUp;
  end;

  { Allocate the AV-FormatContext }
  aFormatCtx := avformat_alloc_context();
  if (aFormatCtx = nil) then
    begin
    Log('! FF-Err: Cannot allocate the AV-FormatContext.');
    ExitCode := -1;
    GoTo CleanUp;
  end;


  { Open the URL }
  if ( avformat_open_input(@aFormatCtx, PChar(aURL), nil, nil) < 0 ) then
  begin
    Log( sLineBreak
       + '! Cannot open:'
       + sLineBreak
       + sLineBreak
       + '  ' + ExtractFilePath(aURL)
       + sLineBreak
       + '  ' + ExtractFileName(aURL)
       );
    ExitCode := -1;
    GoTo CleanUp;
  end
  else
    Log(sLineBreak + '> Opening...' + sLineBreak);

  { Get StreamInfos }
  if ( avformat_find_stream_info(aFormatCtx, nil) < 0 ) then
  begin
    Log('! FF-Err: Cannot find StreamInformation.');
    ExitCode := -1;
    GoTo CleanUp;
  end;

  { Show StreamInfos }
  Log('> StreamInfos...');
  av_dump_format(aFormatCtx, 0, PChar(aURL), 0);


  { Find FIRST Streams } {$<- Note: Better use 'av_find_best_stream()'. }
  aStreamAIdx := -1;
  aStreamVIdx := -1;
  for aIdx := 0 to pred(aFormatCtx^.nb_streams) do
    if      (aFormatCtx^.streams[aIdx]^.codecpar^.codec_type = AVMEDIA_TYPE_AUDIO) and (aStreamAIdx < 0) then
      aStreamAIdx := aIdx
    else if (aFormatCtx^.streams[aIdx]^.codecpar^.codec_type = AVMEDIA_TYPE_VIDEO) and (aStreamVIdx < 0) then
      aStreamVIdx := aIdx;

  { Check, IF Streams were FOUND }
  if ( (aStreamAIdx < 0) and (aStreamVIdx < 0) ) then
  begin
    Log(sLineBreak + '! NO Video- and NO AudioStream found.');
    ExitCode := -1;
    GoTo CleanUp;
  end;

  Pff^.StreamAIdx := aStreamAIdx;
  Pff^.StreamVIdx := aStreamVIdx;

  { IF the AudioStream is VALID...Open the AUDIO-Component }
  if (aStreamAIdx >= 0) then
    if (ff_OpenComponents(aStreamAIdx, aFormatCtx) < 0) then
    begin
      Log(sLineBreak + '! Cannot open AudioComponent.');
      ExitCode := -1;
      GoTo CleanUp;
    end;

  { IF the VideoStream is VALID...Open the VIDEO-Component }
  if (aStreamVIdx >= 0) then
    if (ff_OpenComponents(aStreamVIdx, aFormatCtx) < 0) then
    begin
      Log(sLineBreak + '! Cannot open VideoComponent.');
      ExitCode := -1;
      GoTo CleanUp;
    end;

  Pff^.FormatCtx := aFormatCtx;


  { Open the SDL-Window... }

  if True then {$<- Do it always}
//if (Pff^.CodecCtxV = nil) then
  begin
    aSdlWndW := 640;
    aSdlWndH := 480;
  end
  else
  begin
    aSdlWndW := Pff^.CodecCtxV^.width;
    aSdlWndH := Pff^.CodecCtxV^.height;
  end;

  { Create the SDL-Window }
  aSdlWindow := SDL_CreateWindow(PChar(ExtractFileName(aURL)),
                                 SDL_WINDOWPOS_UNDEFINED,
                                 SDL_WINDOWPOS_UNDEFINED,
                                 aSdlWndW,
                                 aSdlWndH,
                                 SDL_WINDOW_RESIZABLE);
  { Check it }
  if (aSdlWindow = nil) then
  begin
    Log( Format('! SDL-Err: Cannot create SDL-Window (SDL-Err: %s).', [SDL_GetError]) );
    ExitCode := -1;
    GoTo CleanUp;
  end
  else
    Pff^.SdlWindow := aSdlWindow;

  { Create the 2D-RenderingContext for the SDL-Window }
  aSdlRenderer := SDL_CreateRenderer(Pff^.SdlWindow, -1, SDL_RENDERER_ACCELERATED or SDL_RENDERER_PRESENTVSYNC);
  if (aSdlRenderer = nil) then
  begin
    Log( Format('! SDL-Err: Cannot initialize a HW-accelerated SDL-Renderer (SDL-Err: %s).', [SDL_GetError]) );
    ExitCode := -1;
    aSdlRenderer := SDL_CreateRenderer(Pff^.SdlWindow, -1, 0);
  end;

  { Check it }
  if (aSdlRenderer = nil) then
  begin
    Log( Format('! SDL-Err: Cannot create SDL-Renderer (SDL-Err: %s).', [SDL_GetError]) );
    ExitCode := -1;
    GoTo CleanUp;
  end
  else
    Pff^.SdlRenderer := aSdlRenderer;

  { IF it's Audio, only...Draw a BLACK Screen } {nicer}
  if (aStreamVIdx < 0) then
    if (SDL_SetRenderDrawColor(aSdlRenderer, 0, 0, 0, 255) < 0) then
      Log( Format('! SDL-Err: Cannot set SDL-RenderDrawColor (SDL-Err: %s).', [SDL_GetError]) )
    else
    begin
      SDL_RenderFillRect(aSdlRenderer, nil);
      SDL_RenderPresent (aSdlRenderer);
    end;

  { Start the READ-Thread }
  Pff^.ReadThreadID := SDL_CreateThread(@thread_Read, 'READ-Thread', nil);
  if (Pff^.ReadThreadID = nil) then
  begin
    Log( Format('! SDL-Err: Cannot start the READ-Thread (SDL-Err: %s).', [SDL_GetError]) );
    ExitCode := -1;
    GoTo CleanUp;
  end;


  { Start the SDL-EventLoop... }

  while True do
  begin
    SDL_WaitEvent(@aSdlEvent);

    case aSdlEvent.type_ of

      { Stop via <Esc>... }
      SDL_KEYDOWN:
        case aSdlEvent.key.keysym.sym of
          SDLK_ESCAPE:
            Pff^.QuitRead := True;
        end;
      FF_QUIT_EVENT:
      Log(sLineBreak + '> STOPPED - To exit, click [X].');

      { Close via [X]... }
      SDL_QUITEV:
      begin
        Pff^.QuitRead := True;

        SDL_CondSignal(Pff^.PktQA.SdlCond);
        SDL_CondSignal(Pff^.PktQV.SdlCond);
        SDL_CondSignal(Pff^.picQ_SdlCond);

        { Close the SDL-Window... }
        if (Pff^.SdlRenderer <> nil) then
          SDL_DestroyRenderer(Pff^.SdlRenderer);
        if (Pff^.SdlWindow <> nil) then
          SDL_DestroyWindow(Pff^.SdlWindow);

        SDL_Quit();  // Close SDL
        Break;
      end;


      FF_ALLOC_EVENT:
        picQ_AllocPic();  // Allocate a NEW SDL-VideoImage

      FF_REFRESH_EVENT:
        video_Refresh();  // Refresh the SDL-VideoImage

    end;
  end;


CleanUp:
  Log(sLineBreak + '> Cleaning up...');

  { CleanUp the FFPlayer-Threads }
  if (Pff^.VideoThreadID <> nil) then SDL_WaitThread(Pff^.VideoThreadID, nil);
  if (Pff^.ReadThreadID  <> nil) then SDL_WaitThread(Pff^.ReadThreadID , nil);

  { CleanUp the AudioBuffers }
  if (Pff^.AudioBuf   <> nil) then FreeMem(Pff^.AudioBuf  , MAX_AUDIOBUFFER_SIZE);
  if (Pff^.AudioBufRs <> nil) then FreeMem(Pff^.AudioBufRs, MAX_AUDIOBUFFER_SIZE);

  { CleanUp FFmpeg }
  if (Pff^.FrameVOut <> nil) then av_frame_free(@Pff^.FrameVOut);
  if (Pff^.FrameA    <> nil) then av_frame_free(@Pff^.FrameA   );
  if (Pff^.CodecCtxA <> nil) then avcodec_free_context(@Pff^.CodecCtxA);
  if (Pff^.CodecCtxV <> nil) then avcodec_free_context(@Pff^.CodecCtxV);
  if (Pff^.SwrCtx <> nil) then swr_free(@Pff^.SwrCtx);
  if (Pff^.SwsCtx <> nil) then sws_freeContext(Pff^.SwsCtx);

  { Close the URL }
  if (Pff^.FormatCtx <> nil) then avformat_close_input(@Pff^.FormatCtx);

  av_freep(@Pff);  // Free&Nil the FFPlayer-Memory

  { Close the FFmpeg-NetworkComponents }
  avformat_network_deinit();

  Log(sLineBreak + '> CleanUp DONE.' + sLineBreak);

  if (ExitCode < 0) then ReadLn();  // Keep the Terminal opened 'OnError'
end;


{*** Program ******************************************************************}

{$R *.res}

begin

  try
    ExitCode := ff_main();
  except
    on E: Exception do
      WriteLn(E.ClassName, ': ', E.Message);
  end;

(*** END **********************************************************************)

end.

